/***************************************************************************
 *   Copyright (C) 2006 by Gerd v. Egidy                                   *
 *   gve@intra2net.com                                                     *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License version   *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>

#include "socket_server.hxx"
#include "t2n_exception.hxx"

using namespace std;

namespace libt2n
{

socket_server::socket_server(int port, const char* ip)
    : server()
{
    socket_type=tcp_s;

}

socket_server::socket_server(const char* path, mode_t filemode, const char* user, const char* group)
    : server()
{
    socket_type=unix_s;
    unix_path=path;

    /* Create the socket. */
    sock = socket (PF_UNIX, SOCK_STREAM, 0);
    if (sock < 0)
    {
        string err="error opening socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    set_socket_options(sock);

    /* Give the socket a name. */
    struct sockaddr_un unix_name;
    unix_name.sun_family = AF_UNIX;
    strncpy (unix_name.sun_path, unix_path.c_str(),sizeof(unix_name.sun_path));

    /* just to make sure there is no other socket file */
    unlink (unix_name.sun_path);

    if (bind (sock, (struct sockaddr *) &unix_name, sizeof (unix_name)) < 0)
    {
        string err="error binding socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    /* change permissions */
    if (chmod (unix_name.sun_path, filemode) != 0) 
    {
        string err="error changing permission: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    struct passwd *socket_user = getpwnam (user);
    if (socket_user == NULL) 
    {
        string err="error getting socket user: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    struct group *socket_group = getgrnam (group);
    if (socket_group == NULL) 
    {
        string err="error getting socket group: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    if (chown (unix_name.sun_path, socket_user->pw_uid, socket_group->gr_gid) != 0) 
    {
        string err="error changing socket ownership: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    if (listen (sock, 5) < 0)
    {
        string err="error listening to socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    /* clear & insert server sock into the fd_tab to prepare select */
    FD_ZERO(&connection_set);
    FD_SET (sock, &connection_set);
}

void socket_server::set_socket_options(int sock)
{
    int i=1;

    /* fast reuse enable */
    if (setsockopt(sock,SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0)
    {
        string err="error setting socket option: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    /* keepalive enable */
    if (setsockopt(sock,SOL_SOCKET, SO_KEEPALIVE, &i, sizeof(i)) < 0)
    {
        string err="error setting socket option: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    /* close on exec */
    int fdflags;
    fdflags=fcntl(sock,F_GETFD, 0);
    if (fdflags < 0)
    {
        string err="fcntl error on socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }
    fdflags |= FD_CLOEXEC;
    if (fcntl(sock,F_SETFD,fdflags) < 0)
    {
        string err="fcntl error on socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    /* non-blocking mode */
    int flflags;
    flflags=fcntl(sock,F_GETFL,0);
    if (flflags < 0)
    {
        string err="fcntl error on socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }
    flflags |= O_NONBLOCK;
    if (fcntl(sock,F_SETFL,flflags) < 0)
    {
        string err="fcntl error on socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }
}

socket_server::~socket_server()
{
    close(sock);

    if (socket_type==unix_s)
        unlink(unix_path.c_str());
}

void socket_server::new_connection()
{
    struct sockaddr_un clientname;

    unsigned int size = sizeof (clientname);
    int newsock = accept (sock,(struct sockaddr *) &clientname,&size);
    if (newsock < 0)
    {
        if (errno == EAGAIN)
        {
            log(error, "accept error (EAGAIN): no connection waiting");
            return;
        }

        /* default: break */
        string err="error accepting connection: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    FD_SET (newsock, &connection_set);

    int i=1;

    /* keepalive enable */
    if (setsockopt(newsock,SOL_SOCKET, SO_KEEPALIVE, &i, sizeof(i)) < 0)
    {
        string err="error setting socket option: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    /* close on exec */
    int fdflags;
    fdflags=fcntl(newsock,F_GETFD, 0);
    if (fdflags < 0)
    {
        string err="fcntl error on socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }
    fdflags |= FD_CLOEXEC;
    if (fcntl(newsock,F_SETFD,fdflags) < 0)
    {
        string err="fcntl error on socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    /* non-blocking mode */
    int flflags;
    flflags=fcntl(newsock,F_GETFL,0);
    if (flflags < 0)
    {
        string err="fcntl error on socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }
    flflags |= O_NONBLOCK;
    if (fcntl(newsock,F_SETFL,flflags) < 0)
    {
        string err="fcntl error on socket: ";
        err+=strerror(errno);
        log(error, err);
        throw t2n_server_error(err);
    }

    add_connection(new socket_connection(newsock, get_default_timeout()));

    return;
}

void socket_server::fill_buffer(long long usec_timeout)
{
    fd_set used_fdset=connection_set;

    /* set timeout */
    struct timeval tval;
    struct timeval *timeout_ptr;

    if (usec_timeout == -1)
        timeout_ptr = NULL;
    else
    {
        timeout_ptr = &tval;

        // timeout von long long usec in int sec + int usec umrechnen
        tval.tv_sec = usec_timeout / 1000000;
        tval.tv_usec = usec_timeout % 1000000;
    }

    int ret=select (FD_SETSIZE, &used_fdset, NULL, NULL, timeout_ptr);

    if (ret < 0)
    {
        if (errno == EINTR)
        {
            // select interrupted by signal
            ret=0;
        }
        else
        {
            string err="select error: ";
            err+=strerror(errno);
            log(error, err);
            throw t2n_server_error(err);
        }
    }

    if (ret > 0)
    {
        // we have data pending

        // check for new connection
        if (FD_ISSET (sock, &used_fdset))
        {
            new_connection();
        }

        // check all connections for pending data
        fill_connection_buffers();
    }

    return;
}

socket_connection::socket_connection(int _sock, int _timeout)
    : connection(_timeout)
{
    sock=_sock;
}

void socket_connection::close()
{

}

void socket_connection::fill_buffer(void)
{

}

void socket_connection::write(const std::string& data)
{

}

}
